/*
 * Virtual client module
 * Copyright 2017, 2018 William Lupshenko
 * 
 * This file is a part of TUServer
 * 
 * TUServer is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Pirate Public License, as
 * published by William Lupshenko. See the LICENSE file for more details
 */

var
	crypto = require("crypto"),
	fs = require("fs"),
	zlib = require("zlib"),
	url = require("url"),
	path = require("path"),
	net = require("net");

module.exports = {
	//Secondary variables
		players: {
			//Secondary variables
				defaultGuiData: {},
				
				list: {},
			
			//Methods
				New: function(name) {
					var spawnPoint = module.exports.players.findSpawnPoint();
					
					Object.assign(this, 
						new game.realms.gameWorld.objectTypes.player.New(),
						
						{
							name: name,
							
							x: String(spawnPoint.x),
							y: String(spawnPoint.y),
							
							cameraX: String(spawnPoint.x.subtract(game.realms.gameWorld.playerOffset.x)),
							cameraY: String(spawnPoint.y.subtract(game.realms.gameWorld.playerOffset.y)),
							
							currentGuiType: "default",
							guiData: JSON.parse(JSON.stringify(
								module.exports.players.defaultGuiData
							))
						}
					);
				},
				
				findSpawnPoint: function() {
					for (
						var
							x = BigInteger.randBetween("-9007199254740991", "9007199254740991"),
							y = BigInteger.randBetween("-9007199254740991", "9007199254740991");;
					) {
						if (!game.realms.gameWorld.objects[[x, y]]) {
							return {
								x: x,
								y: y
							};
						}
					}
				},
				
				setCurrent: function(player) {
					if (this.list[player]) {
						this.current = player;
						
						game.config.list.gameWorld.playerPosition.x =
							this.list[player].x;
						game.config.list.gameWorld.playerPosition.y =
							this.list[player].y;
						
						game.config.list.gameWorld.cameraPosition.x =
							this.list[player].cameraX;
						game.config.list.gameWorld.cameraPosition.y =
							this.list[player].cameraY;
						
						game.realms.gameWorld.currentGuiType.data =
							this.list[player].guiData[
								this.list[player].currentGuiType
							];
						
						if (
							game.config.list.gameWorld.playerPosition.x > 9007199254740991 ||
							game.config.list.gameWorld.playerPosition.x < -9007199254740991 ||
							game.config.list.gameWorld.playerPosition.y > 9007199254740991 ||
							game.config.list.gameWorld.playerPosition.y < -9007199254740991 ||
							 
							game.config.list.gameWorld.cameraPosition.x > 9007199254740991 ||
							game.config.list.gameWorld.cameraPosition.x < -9007199254740991 ||
							game.config.list.gameWorld.cameraPosition.y > 9007199254740991 ||
							game.config.list.gameWorld.cameraPosition.y < -9007199254740991
						) {
							alert(String(new SyntaxError("illegal player/camera position")));
							game.realms.gameWorld.fail(this.list[player]);
						}
					}
				},
				update: function(player) {
					if (this.list[player]) {
						this.list[player].x = game.config.list.gameWorld.playerPosition.x
						this.list[player].y = game.config.list.gameWorld.playerPosition.y;
						
						this.list[player].cameraX = game.config.list.gameWorld.cameraPosition.x
						this.list[player].cameraY = game.config.list.gameWorld.cameraPosition.y;
					}
				},
				
				connect: function(player) {
					Object.defineProperty(this.list, player, {
						configurable: true,
						enumerable: true,
						
						get: new Function("",
							"return game.config.list.players[" +
								"\"" + player + "\"" +
							"];"
						),
						set: new Function("value",
							"game.config.list.players[" +
								"\"" + player + "\"" +
							"] = value;"
						)
					});
					
					if (!game.config.list.players[player]) {
						this.list[player] = new this.New(player);
					}
					
					game.realms.gameWorld.objects[[
						this.list[player].x,
						this.list[player].y
					]] = this.list[player];
					
					console.log(player + " logged in");
					game.realms.gameWorld.guiTypes.default.writeToChat("",
						player + game.currentLocale.gui_multiplayerWorld_loggedIn,
					game.colorScheme.text_good);
				},
				disconnect: function(player) {
					if (player in this.list) {
						this.list[player].currentGuiType = "default";
						
						game.realms.gameWorld.objects[[
							this.list[player].x,
							this.list[player].y
						]] = new game.realms.gameWorld.objectTypes.grave.New({
							name: player
						});
						
						delete this.list[player];
						
						console.log(player + " logged out");
						game.realms.gameWorld.guiTypes.default.writeToChat("",
							player + game.currentLocale.gui_multiplayerWorld_loggedOut,
						game.colorScheme.text_bad);
					}
				}
		},
	
	//Methods
		init: function() {
			function loadScript(path) {
				try {
					new Function(
						"require",
						"nodeCrypto",
						"fs",
						"zlib",
						"url",
						"path",
						"net",
						
						fs.readFileSync(path, "utf-8")
					)(
						require,
						crypto,
						fs,
						zlib,
						url,
						path,
						net
					);
				} catch (error) {
					console.log(new ReferenceError(
						"can't load " + path +
						
						" (" + error + ")"
					));
				}
			}
			
			loadScript("client/BigInteger.js");
			loadScript("client/main.js");
			loadScript("client/main_loadScripts.js");
			loadScript("client/main_loadTexture.js");
			loadScript("client/main_loadSound.js");
			loadScript("client/config.js");
			loadScript("client/hotkeys.js");
			loadScript("client/mods.js");
			
			game.loadScripts();
			
			game.config.read();
			
			//Check if spawn cluster isn't generated
				if (!game.config.list.gameWorld.clusters[[0, 0]]) {
					game.realms.gameWorld.generator.generateSpawnCluster();
					
					game.realms.gameWorld.objects[[
						game.config.list.gameWorld.playerPosition.x,
						game.config.list.gameWorld.playerPosition.y
					]] = null;
					
					game.config.write("gameWorld");
				}
			
			//Generate default GUI data
				for (var i in game.realms.gameWorld.guiTypes) {
					this.players.defaultGuiData[i] =
						game.realms.gameWorld.guiTypes[i].defaultData;
				}
			
			this.updater = setInterval(function() {
				app.virtualClient.updateChunks();
				
				app.virtualClient.render();
			}, 40);
		},
		
		updateChunks: function() {
			for (var i in this.players.list) {
				this.players.setCurrent(i);
				
				game.realms.gameWorld.updateChunk();
				
				this.players.update(i);
			}
			
			game.realms.gameWorld.updatedAreas.length = 0;
			
			if (game.realms.gameWorld.tick < 24) {
				game.realms.gameWorld.tick++;
			} else {
				game.realms.gameWorld.tick = 0;
			}
		},
		
		handleHotkey: function(player, keyCode, isDown) {
			this.players.setCurrent(player);
			
			game.hotkeys.catch(keyCode, isDown);
			
			game.realms.gameWorld.currentGuiType.handleHotkeys();
			
			this.players.update(player);
		},
		
		render: function() {
			for (var i in this.players.list) {
				this.players.setCurrent(i);
				
				if (
					"result" in app.sessions[i].data.prompt &&
					"handler" in app.sessions[i].data.prompt
				) {
					app.sessions[i].data.prompt.handler(
						app.sessions[i].data.prompt.result
					);
					
					delete app.sessions[i].data.prompt.result;
					delete app.sessions[i].data.prompt.handler;
				}
				if (
					"result" in app.sessions[i].data.confirm &&
					"handler" in app.sessions[i].data.confirm
				) {
					app.sessions[i].data.confirm.handler(
						app.sessions[i].data.confirm.result
					);
					
					delete app.sessions[i].data.confirm.result;
					delete app.sessions[i].data.confirm.handler;
				}
				
				if (i in this.players.list) {
					var renderingData = {
						gfx: game.realms.gameWorld.update(),
						
						debugInfo: {
							software:
								app.manifest.name +
								" (" + game.name + " " + game.version + ") " +
								app.manifest.version,
							
							cameraPosition: game.config.list.gameWorld.cameraPosition,
							playerPosition: game.config.list.gameWorld.playerPosition
						}
					};
					
					if (app.sessions[i].data.buffer != JSON.stringify(renderingData)) {
						app.sessions[i].data.buffer = JSON.stringify(renderingData);
						
						app.sessions[i].reply("rendering", renderingData);
					}
				}
				
				this.players.update(i);
			}
		}
};
